use crate::config;
use clap::{ArgAction, Args, Parser};
use eyre::ensure;
use futures::try_join;
use listenfd::ListenFd;
use std::{net::TcpListener, os::unix::net::UnixListener, path::PathBuf};

#[derive(Parser, Debug)]
pub struct CLI {
    #[arg(long, value_parser, value_name = "path/to/config.toml")]
    pub config: config::Config,
    #[command(flatten)]
    pub listen: Listen,
}

impl CLI {
    pub async fn resolve(&mut self, listenfd: &mut ListenFd) -> eyre::Result<()> {
        try_join!(self.config.resolve(), self.listen.resolve(listenfd))?;
        Ok(())
    }
}

#[derive(Args, Debug)]
#[group(required = true)]
pub struct Listen {
    #[arg(long, value_name = "HOST:PORT")]
    pub listen_tcp: Vec<String>,
    #[arg(long, value_name = "path/to/socket.sock")]
    pub listen_unix: Vec<PathBuf>,
    #[arg(long, action = ArgAction::SetTrue, default_missing_value = "", value_parser = |_: &str| Ok::<_, String>(ListenFdSockets(vec![])))]
    pub listen_fd: Option<ListenFdSockets>,
}

#[derive(Debug)]
pub struct ListenFdSockets(pub Vec<ListenFdSocket>);

impl Clone for ListenFdSockets {
    fn clone(&self) -> Self {
        panic!("not clonable")
    }
}

#[derive(Debug)]
pub enum ListenFdSocket {
    Tcp(TcpListener),
    Unix(UnixListener),
}

impl Listen {
    pub async fn resolve(&mut self, listenfd: &mut ListenFd) -> eyre::Result<()> {
        if let Some(listen_fd_sockets) = &mut self.listen_fd {
            ensure!(
                listenfd.len() != 0,
                "no file descriptor passed in through LISTEN_FDS systemd protocol"
            );
            for index in 0..listenfd.len() {
                let mut result = listenfd
                    .take_tcp_listener(index)
                    .transpose()
                    .expect("fds not yet taken")
                    .map(ListenFdSocket::Tcp);
                if result.is_err() {
                    if let Ok(Some(v)) = listenfd.take_unix_listener(index) {
                        result = Ok(ListenFdSocket::Unix(v));
                    }
                }
                listen_fd_sockets.0.push(result?);
            }
        }
        Ok(())
    }
}
